#####################################################################
# Copyright (C) 2000 Artur Trzewik
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
######################################################################
# Header
# CVS - $Id: formfactory.tcl,v 1.9 2000/10/15 17:19:20 artur Exp $
#
# Information:
# Modul (benutzt von Form Editor als auch Form Server)
# Entwurfsmuster Fabrik
# Produzier widgetlib Struktur aus XML tDOM Represenatation von DB Formularen
#
# Die sqlconnect und hidedata werden nur von FormServer bei realen Einsatz 
# benutzt 
# defaults ist ein Array mit gesammelten default Werten 
# gebracht bei Aktion neues Tupel genau so notnulls
Class FormFactory -parameter {
    sqlconnect
    {hidedata {}}
    default_arr
    notnull_arr
}
FormFactory instproc makedialog_fromdom {dom_node {parent {}}} {
    [self] instvar data editable notnull_data def_values
    set notnull_data {}
    set def_values {}
    set editable 0
    set data {}
    foreach childnode [$dom_node childNodes] {
	if {[$childnode nodeType]=="ELEMENT_NODE" && 
	    [$childnode nodeName]=="frame"} {
	    set name [[self] makedom_frame $childnode $parent]
	    break
	}
    }
#     foreach childnode [$dom_node childNodes] {
# 	set wname [[self] makedom_[$childnode nodeName] $childnode $name]
# 	$name addwidget $wname {}
#     }
    return $name
}
# Widget Factory fr DBGuiEditor
# Es werden andere (flache) Frame Widget erzeugt 
FormFactory instproc makeeditable_fromdom {dom_node guiedit} {
    [self] instvar data editable
    set editable 1
    set rootframe {}
    if {[info exist data]} { unset data }
    foreach childnode [$dom_node childNodes] {
	if {[$childnode nodeType]=="ELEMENT_NODE" && 
	    [$childnode nodeName]=="frame"} {
	    set rootframe [[self] makedom_frame $childnode $guiedit]
	    break
	}
    }
    return $rootframe
}
FormFactory instproc makedom_frame {frame_node parent} {
    #puts "makedom_frame $frame_node $parent"
    [self] make_frame_[$frame_node getAttribute manager] $frame_node $parent
}
FormFactory instproc make_frame_pack {frame_node parent} {
    [self] instvar editable
    if {$editable} {
	set wclass EPackFrame
	set par [[self] extParameter $wclass $frame_node]
	set packer [eval EPackFrame ${parent}::[Object autoname myEPackFrame] $par]
	set parentc $parent
    } else {
	set wclass FrameContainer
	set par [[self] extParameter $wclass $frame_node]
	#puts "par $par"
	if {$parent==""} {
	    set parentc [eval FrameContainer [Object autoname myFrameContainer] $par]
	} else {
	    set parentc [eval FrameContainer ${parent}::[Object autoname myFrameContainer] $par]
	}
	set packer $parentc
    }
    foreach pack [$frame_node childNodes] {
	if {[$pack nodeType]=="ELEMENT_NODE" && 
	    [$pack nodeName]=="pack"} {
	    set packoptions {}
	    foreach opt [$pack attributes] {
		lappend packoptions -$opt [$pack getAttribute $opt]
	    }
	    foreach elem [$pack childNodes] {
		if {[$elem nodeType]=="ELEMENT_NODE"} {
		    #puts "packing by [self] [$elem nodeName]"
		    set widget [makedom_[$elem nodeName] $elem $parentc]
		    if {$widget==""} continue
		    $packer addwidget $widget $packoptions
		}
	    }
	}	
    }
    return $packer
}
# Interne Prozeduren (Erzeuger) der Klassen fr den Aufbau 
# der Dialogen aus dom (XML) Schnittstelle
# makedom_{elementname} {dom_node parent editable}
# wenn editable 1 dann wird eine EditWidget Klasse gebildet
# (fr DBGuiEditor} auch data wird anders interpretiert 
# statt einer Liste nun ein Array 
# um die Zuordnung data->widget zu gewhrleisten
# Rckgabewert Klassenname
# Vorsicht andere Bedeutung von default bei Widget und Formular Elementen

# Hilfsprozedur Erzeugt einem Name und initialisiert data 
FormFactory instproc makeName {dom_node parent} {
    [self] instvar data editable
    set adata [$dom_node getAttribute data]
    #Diese Methode ist zuverlssig weil keine Vordopplungen vorkommen knnen
    #Aber es verhinder die Benutzung von X Configure Options
    #Weil die Name von Widget nicht bestimmbar sind
    #Tk verbitet die Benutzung der ersten groen Buchstaben
    if {$editable} {
	set name [Object autoname myWidget]
	lappend data($adata) ${parent}::$name
    } else {
	#set name [string tolower $adata]
	set name [Object autoname myWidget]
	if {![lcontain $data $adata]} {
	    lappend data $adata
	}
    }
    return $name
}
# Diese Prozedur extrahiert aus XML die Widget parameter und wandelt es in eine Liste
# es wird die metadata des Widget-Klassen benutzt
FormFactory instproc extParameter {class node} {
    set domattrs [$node attributes]
    set attr [$class metadata attributes]
    set par {}
    foreach domattr [schnittMenge $domattrs $attr] {
	    lappend par -$domattr [$node getAttribute $domattr] 
    }
    return $par
}
# Initiert die Datenhrekunft Relevante Attribute wie
# data, notnull, default
# Unterschiedliche Behanldung bei EditFrame (GUI Editor) und 
# FormEngine (Form Server)
# bie EditFrame gehen alle Attribute als Keyed List in Attribut data
# bie FormEngine werden die dafault und notnull in spezielle Keyed List
# des Objekt [self] refasst in data geht nur Eigenschaft data
FormFactory instproc setDataSource {object dom_node} {
    [self] instvar editable def_values notnull_data
    set domattrs [$dom_node attributes]
    # Extrahiere Datenherkunft relevante Attribute
    foreach dataattr [schnittMenge $domattrs {data default notnull}] {
	keylset datakl $dataattr [$dom_node getAttribute $dataattr] 
    }
    if {$editable} {
	$object set data $datakl
    } else {
	$object set data [keylget datakl data]
	if {[lcontain $domattrs notnull]} {
	    lappend notnull_data  [keylget datakl data]
	}
	if {[lcontain $domattrs default]} {
	    keylset def_values [keylget datakl data] [keylget datakl default]
	}
    }
}
FormFactory instproc makedom_text {dom_node parent} {
    [self] instvar data editable hidedata
    # Wenn zu Hidedate dehrt dann ignoriere (Wichtig bei eigebeteten formularen)
    if {[lcontain $hidedata [$dom_node getAttribute data]]} {
	return
    }
    set name [[self] makeName $dom_node $parent]
    set domattrs [$dom_node attributes]
    # Bestimme Widget Klasse
    if {[lcontain $domattrs subtype] && [$dom_node getAttribute subtype]=="area"} {
	set wclass OLabelText
    } else {
	set wclass OLabelEntry
    }
    set par [[self] extParameter $wclass $dom_node]
    if {$editable} {
	eval E${wclass} ${parent}::$name $par
    } else {
	eval $wclass ${parent}::$name $par
    }
    [self] setDataSource ${parent}::$name $dom_node 
    return ${parent}::$name
}
FormFactory instproc makedom_boolean {dom_node parent} {
    [self] instvar data editable hidedata
    # Wenn zu Hidedate dehrt dann ignoriere (Wichtig bei eigebeteten formularen)
    if {[lcontain $hidedata [$dom_node getAttribute data]]} {
	return
    }
    set wclass OCheckButton
    set name [[self] makeName $dom_node $parent]
    set par [[self] extParameter $wclass $dom_node]
    if {$editable} {
	eval E${wclass} ${parent}::$name $par
    } else {
	eval $wclass ${parent}::$name $par
    }
    [self] setDataSource ${parent}::$name $dom_node 
    return ${parent}::$name
}
FormFactory instproc makedom_numeric {dom_node parent} {
    [self] instvar data editable hidedata
    # Wenn zu Hidedate dehrt dann ignoriere (Wichtig bei eigebeteten formularen)
    if {[lcontain $hidedata [$dom_node getAttribute data]]} {
	return
    }
    set wclass OLabelControl
    set name [[self] makeName $dom_node $parent]
    set par [[self] extParameter $wclass $dom_node]
    if {$editable} {
	eval E${wclass} ${parent}::$name $par
    } else {
	eval $wclass ${parent}::$name $par
    }
    [self] setDataSource ${parent}::$name $dom_node 
    return ${parent}::$name
}
FormFactory instproc makedom_list {dom_node parent} {
    [self] instvar data editable hidedata 
    # Wenn zu Hidedate dehrt dann ignoriere (Wichtig bei eigebeteten formularen)
    if {[lcontain $hidedata [$dom_node getAttribute data]]} {
	return
    }
    set name [[self] makeName $dom_node $parent]
    set domattrs [$dom_node attributes]
    # Bestimme Widget Klasse
    if {[lcontain $domattrs subtype] && [$dom_node getAttribute subtype]=="radiobuttons"} {
	set wclass ORadioButtons
    } else {
	set wclass OComboBox
    }
    set par [[self] extParameter $wclass $dom_node]
    lappend par -valuelist [tdom_getvaluelist $dom_node]
    if {$editable} {
	eval E${wclass} ${parent}::$name $par
    } else {
	eval $wclass ${parent}::$name $par
    }
        [self] setDataSource ${parent}::$name $dom_node 
    return ${parent}::$name
}
FormFactory instproc makedom_reference {dom_node parent} {
    [self] instvar data editable sqlconnect hidedata
    # Wenn zu Hidedate dehrt dann ignoriere (Wichtig bei eigebeteten formularen)
    if {[lcontain $hidedata [$dom_node getAttribute data]]} {
	return
    }
    set name [[self] makeName $dom_node $parent]
    # Bestimme Widget Klasse
    set wclass OSqlReferenceList
    set par [[self] extParameter $wclass $dom_node]
    lappend par -sqlquery [tdom_getpername $dom_node sqlquery]
    if {$editable} {
	eval E${wclass} ${parent}::$name $par
    } else {
	eval $wclass ${parent}::$name $par -sqlconnect [list $sqlconnect]
    }
    [self] setDataSource ${parent}::$name $dom_node 
    return ${parent}::$name
}
FormFactory instproc makedom_nestedform {dom_node parent} {
    [self] instvar data editable sqlconnect
    set name [[self] makeName $dom_node $parent]
    set domattrs [$dom_node attributes]
    # Bestimme Widget Klasse
    if {$editable} {
	set wclass EFormEngine
    } else {
	set wclass FormEngine
    }
    set attr [$wclass metadata attributes]
    # Extrahiere Widgeteigene Attribute wie textwidth
    foreach domattr [schnittMenge $domattrs $attr] {
	if {!$editable && $domattr=="form"} {
	    lappend par -form_node [tdom_findpername $dom_node form [$dom_node getAttribute form]]
	} else {
	    lappend par -$domattr [$dom_node getAttribute $domattr] 
	}
    }
    # Extrahiere Datenherkunft relevante Attribute
    if {$editable} {
	eval ${wclass} ${parent}::$name $par
    } else {
	eval $wclass ${parent}::$name $par -nestedform 1 -sqlconnect [list $sqlconnect]
    }
        [self] setDataSource ${parent}::$name $dom_node 
    return ${parent}::$name
}
FormFactory instproc makedom_formlink {dom_node parent} {
    [self] instvar data editable sqlconnect
    set name [[self] makeName $dom_node $parent]
    set domattrs [$dom_node attributes]
    # Bestimme Widget Klasse
    if {$editable} {
	set wclass EFormLink
    } else {
	set wclass FormLink
    }
    set attr [$wclass metadata attributes]
    # Extrahiere Widgeteigene Attribute wie textwidth
    foreach domattr [schnittMenge $domattrs $attr] {
	if {!$editable && $domattr=="form"} {
	    lappend par -form_node [tdom_findpername $dom_node form [$dom_node getAttribute form]]
	} else {
	    lappend par -$domattr [$dom_node getAttribute $domattr] 
	}
    }
    if {$editable} {
	eval ${wclass} ${parent}::$name $par
    } else {
	# Vorsicht ein bser Hack bei bergabe von sqlconnect !!!
	eval $wclass ${parent}::$name $par -sqlconnect [list $sqlconnect]
    }
    [self] setDataSource ${parent}::$name $dom_node 
    return ${parent}::$name
}
FormFactory instproc makedom_inheritform {dom_node parent} {
    [self] instvar data editable sqlconnect
    set name [[self] makeName $dom_node $parent]
    set domattrs [$dom_node attributes]
    # Bestimme Widget Klasse
    if {$editable} {
	set wclass EInheritForm
    } else {
	set wclass InheritForm
    }
    set attr [$wclass metadata attributes]
    # Extrahiere Widgeteigene Attribute wie textwidth
    set par {}
    foreach domattr [schnittMenge $domattrs $attr] {
	lappend par -$domattr [$dom_node getAttribute $domattr] 
    }
    if {$editable} {
	set list_label {}; set list_form {}; set list_inheritkey {}
	foreach child [$dom_node selectNodes child::child] {
	    foreach attr {label form inheritkey} {
		if {[$child hasAttribute $attr]} {
		    lappend list_$attr [$child getAttribute $attr]
		} else {
		    lappend list_$attr {}
		}
	    }
	}
	keylset childforms label $list_label
	keylset childforms form $list_form
	keylset childforms inheritkey $list_inheritkey
	eval ${wclass} ${parent}::$name $par -childforms [list $childforms]
    } else {
	eval $wclass ${parent}::$name $par
	# Die Unterstruktur des InheritForm bilden
	set children [$dom_node selectNodes child::child]
	if {[llength $children]>1} {
	    set notebook [NoteBookContainer ${parent}::${name}::notebook]
	    foreach child $children {
		set formName [$child getAttribute form]
		set newcard [FormEngine ${notebook}::$formName -nestedform 1 -data [$dom_node getAttribute data] -sqlconnect $sqlconnect -form_node [$child selectNodes //child::form\[@name=\"$formName\"\]] -refattribute [$child getAttribute inheritkey] -oneobject 1]
		$notebook addcard $newcard [list $formName [$child getAttribute label]]
	    }
	} else {
	    set formName [$children getAttribute form]
	    FormEngine ${parent}::${name}::form -nestedform 1 -sqlconnect $sqlconnect -form_node [$children selectNodes //child::form\[@name=\"$formName\"\]] -refattribute [$children getAttribute inheritkey] -oneobject 1 -border 1
	}
    }
    # Behandle Datenhrekungf Attribute data, default, notnull
    [self] setDataSource ${parent}::$name $dom_node 
    return ${parent}::$name
}
